
#include <QMessageBox>
#include <QFileDialog>
#include <QSplashScreen>
#include <QTextCursor>

#include <math.h>
#include "CorteoParamWidget.h"
#include "constants.h"

CorteoParamWidget::CorteoParamWidget(QWidget * parent, int maxlayer, int maxelement, QSplashScreen * splash)
                  : QWidget(parent)
{
  // setup base class
  setupUi(this);

  // icons indicating if a layer or an element is used (green) or not (gray)
  green = QIcon(":/images/green.png");
  gray  = QIcon(":/images/gray.png");

  // validators for numerical edit fields
  FSDoubleValidator * fsdValidator =    new FSDoubleValidator(this);
  FSDoubleValidator * fsdPosValidator = new FSDoubleValidator(this, 0.0); // validator positive parameters
  QIntValidator     * intValidator =    new QIntValidator(this);
  intValidator->setBottom(0);

  // set technique items
  techniqueName  << "RBS"    << "RBS-TOF"    << "ERD"    << "ERD-TOF"  << "Coincidence" << "Full RBS" << "Full RBS background" << "(Implantation)" << "(Full cascade)"  ;
  techniqueCode << TECH_RBS << TECH_RBS_TOF << TECH_ERD << TECH_ERD_TOF << TECH_COINC  << TECH_FULLRBS << TECH_FULLBACK << TECH_IMPLANT  << TECH_IMPREC  ;
  comboBoxTechnique->addItems(techniqueName);
  comboBoxTechnique->setCurrentIndex(0);

  comboBoxScreening->addItem("None");
  comboBoxScreening->addItem("Andersen");
  comboBoxScreening->addItem("Universal");
  comboBoxScreening->addItem("Lenz-Jensen");

  comboBoxNthread->addItem("1");
#ifdef MAX_THREAD
  for(int ithread=2; ithread<=MAX_THREAD; ithread +=2)
      comboBoxNthread->addItem(QString().setNum(ithread));
#endif


  if(splash) splash->showMessage("setting up layer forms...",Qt::AlignHCenter|Qt::AlignBottom);

  // set layer tabs
  tabWidgetLayers->removeTab(0); // remove default layer tab
  for(int ilayer=0;ilayer<maxlayer;ilayer++) {

    // create layer parameter forms
    LayerForm * lf;
    tabWidgetLayers->addTab(lf = new LayerForm(this, maxelement), "");
    layer.append(lf);

    // connect layer move buttons and useLayer checkbox
    connect( layer[ilayer]->pushButtonMoveToPrevious,   SIGNAL(clicked()), SLOT(moveLayerToPrevious()) );
    connect( layer[ilayer]->pushButtonMoveToNext,       SIGNAL(clicked()), SLOT(moveLayerToNext()) );
    connect( layer[ilayer]->checkBoxUseLayer,           SIGNAL(clicked()), SLOT(useLayerChanged()) );

    // connect element button and checkboxes for update
    for(int ielement=0;ielement<maxelement;ielement++) {
      connect(layer[ilayer]->element[ielement]->checkBoxUseElement, SIGNAL(clicked()), SLOT(update()) );
      connect(layer[ilayer]->element[ielement]->checkBoxSimulate,   SIGNAL(clicked()), SLOT(update()) );
      connect(layer[ilayer]->element[ielement]->checkBoxIsotopes,   SIGNAL(clicked()), SLOT(update()) );
    }
  }

  // create dialog to edit detector foil
  foilDialog = new QDialog(this);
  foil = new LayerForm(foilDialog, maxelement);

  // hide "move" buttons for foils
  foil->pushButtonMoveToPrevious->hide();
  foil->pushButtonMoveToNext->hide();
  connect( pushButtonEditFoil, SIGNAL(clicked()), SLOT(editFoil()) );

  if(splash) splash->showMessage("setting up detector forms...",Qt::AlignHCenter|Qt::AlignBottom);

  // set detector tabs
  tabWidgetDetectors->removeTab(0); // remove default detector tab
  DetectorForm * df;

  // create detectors with specific names and include them in detector tabs
  tabWidgetDetectors->addTab(df = new DetectorForm("RBS", this), "RBS");
  detector.append(df);

  tabWidgetDetectors->addTab(df = new DetectorForm("ERD", this), "ERD");
  detector.append(df);

  tabWidgetDetectors->addTab(df = new DetectorForm("RBS-TOF", this), "RBS-TOF");
  df->labelDistance->setText("Distance from 1st [mm]");
  detector.append(df);

  tabWidgetDetectors->addTab(df = new DetectorForm("ERD-TOF", this), "ERD-TOF");
  df->labelDistance->setText("Distance from 1st [mm]");
  detector.append(df);

  // Calibration forms
  CalibrationForm * cf;
  if(splash) splash->showMessage("setting up calibration forms...",Qt::AlignHCenter|Qt::AlignBottom);

  // at index 0, set a TOF calibration form
  tabWidgetCalibrations->addTab(cf = new CalibrationForm(this, 0, maxlayer), "");
  cf->labelSlope->setText("Quad [ns^2/ch]:");
  cf->labelSlope->setText("Slope [ns/ch]:");
  cf->labelZero->setText("Zero [ns]:");
  cf->labelResolution->setText("Resolution [ns]:");
  cf->labelSlope->setText("Quad (disabled):");
  cf->labelQuad->setEnabled(false);
  cf->lineEditQuad->setText("0"); // quadtratic calibration term not implemented in TOF
  cf->lineEditQuad->setEnabled(false);
  calibration.append(cf);

  // then, append a calibration form for each element in the periodic table
  PeriodicTable periodicTable;
  for(int Z=1; Z<=periodicTable.countAtoms(); Z++) {
    tabWidgetCalibrations->addTab(cf = new CalibrationForm(this, Z, maxlayer), "");
    calibration.append(cf);
  }
  // now remove them all, update()  will put back the useful ones
  while(tabWidgetCalibrations->count()) tabWidgetCalibrations->removeTab(0);

  // verify if move button of tab 0 should be enabled
  moveLayerEnable(0);

  // set validators for toolbox numerical edit fields
  lineEditNumberProjectiles->setValidator(intValidator);
  lineEditRandomSeed->setValidator(intValidator);
  lineEditNumberChannels->setValidator(intValidator);
  lineEditEnergyDiffDisc->setValidator(fsdPosValidator);

  lineEditMeanFreePath->setValidator(fsdValidator);
  lineEditConeAngle->setValidator(fsdPosValidator);
  lineEditMinimumEnergy->setValidator(fsdPosValidator);
  lineEditCutoff->setValidator(fsdPosValidator);
  lineEditEnergy->setValidator(fsdPosValidator);
  lineEditCharge->setValidator(fsdValidator);
  lineEditAngleToNormal->setValidator(fsdPosValidator);
  lineEditAzimAngle->setValidator(fsdValidator);
  lineEditBeamDiameter->setValidator(fsdPosValidator);
  lineEditWfactor->setValidator(fsdPosValidator);
  lineEditHfactor->setValidator(fsdPosValidator);


  // connect a few widgets to their slots
  connect( pushButtonBeam,     SIGNAL(clicked()),                SLOT(setBeam())            );
  connect( tabWidgetLayers,    SIGNAL(currentChanged(int)),      SLOT(moveLayerEnable(int)) );
  connect( pushButtonAbout,    SIGNAL(clicked()),                SLOT(aboutCorteoUi())       );

  connect( comboBoxTechnique,  SIGNAL(currentIndexChanged(int)), SLOT(update())     );
  connect( pushButtonListFile, SIGNAL(clicked()),                SLOT(setlistFile()));

  connect( pushButtonLoad,     SIGNAL(clicked()),                SLOT(loadFile())   );
  connect( pushButtonSaveAs,   SIGNAL(clicked()),                SLOT(saveFileAs()) );

  connect( pushButtonShowCorteoLog,   SIGNAL(clicked()),         SLOT(showCorteoLog()) );

  // try loading last saved parameters
  if(splash) splash->showMessage("loading parameters...",Qt::AlignHCenter|Qt::AlignBottom);
  if( !loadParam("CorteoUi.cml") ) {

    // load last saved params did not work so initialize elements to some default values:
    beamElement = Atom(ElementForm::setElement(2, pushButtonBeam, comboBoxBeamMass)); // beam = He

    // set silicon as default elements in layers
    for(int ilayer=0; ilayer<maxlayer; ilayer++)
      for(int ielement=0; ielement<maxelement; ielement++) {
        layer[ilayer]->element[ielement]->setElement(14);
      }

    // set carbon as default elements in foil
    for(int ielement=0; ielement<maxelement; ielement++)
      foil->element[ielement]->setElement(6);
  }

  update();
}


// determine straggling type from radiobutton state
int CorteoParamWidget::stragglingType() {

    if(radioButtonStragBohr->isChecked()) return 1;
    if(radioButtonStragChu->isChecked()  ) return 2;
    if(radioButtonStragWang->isChecked() ) return 3;

    return 0; // default, if None or no one checked
}


// set straggling selection, or to None if wrong type
void CorteoParamWidget::setStragglingType(int type) {
  switch (type) {
  default:
    QMessageBox::warning(this, "Corteo user interface", "Unknown straggling type, setting it to None.");
    // no break, case 0 applies
  case 0:
    radioButtonStragNone->setChecked(true);
    break;
  case 1:
    radioButtonStragBohr->setChecked(true);
    break;
  case 2:
    radioButtonStragChu->setChecked(true);
    break;
  case 3:
    radioButtonStragWang->setChecked(true);
    break;
  }
}


// call setElement for beam button and mass selection box
void CorteoParamWidget::setBeam() {
  beamElement = ElementForm::selectElement(this, pushButtonBeam, comboBoxBeamMass);
  update();
}


void CorteoParamWidget::useLayerChanged() {
  setUseLayerIcon(tabWidgetLayers->currentIndex());
  update(); // update the list of detected elements
}


void CorteoParamWidget::setUseLayerIcon(int ilayer) {
  if(ilayer>=tabWidgetLayers->count() || ilayer<0) return;

  // icon green or gray whether layer is used or not
  bool bl = layer[ilayer]->isUsed();
  tabWidgetLayers->setTabIcon(ilayer, bl?green:gray);
  layer[ilayer]->labelNumber->setText(QString("Layer #%1").arg(ilayer));

  // change also the icon of all calibration forms compound correction tabs, in case they are used
  for(int i=0; i<calibration.count(); i++)
    calibration[i]->tabWidgetCompCorr->setTabIcon(ilayer, bl?green:gray);
}


// swap forms and container widget of two layers
bool CorteoParamWidget::swapLayers(int layer1, int layer2) {
  if( layer1<0 || layer2<0 || layer1>=layer.count() || layer2>=layer.count() || layer1==layer2 )
    return false;

  // swap layer forms in layer list
  layer.swap(layer1,layer2);

  // insert layer1 at position of layer2 (this does not create a new tab!)
  tabWidgetLayers->insertTab(layer2, tabWidgetLayers->widget(layer1), "");

  // swap compound corrections in calibration forms
  for (int i = 0; i < calibration.count(); i++) {
    // insert widget corresponding to layer1 at position of that of layer2
    // (this does not create a new tab!)
    QWidget * w = calibration[i]->tabWidgetCompCorr->widget(layer1);
    calibration[i]->tabWidgetCompCorr->insertTab(layer2, w, "");
  }

  setUseLayerIcon(layer1);
  setUseLayerIcon(layer2);

  return true;
}


// check if current layer tab can move up or down and enable corresponding buttons
void CorteoParamWidget::moveLayerEnable(int ilayer) {
  if( ilayer<0 || ilayer>=layer.count())
    return; // layer tabs not initialized

  layer[ilayer]->pushButtonMoveToPrevious->setEnabled(ilayer>0);
  layer[ilayer]->pushButtonMoveToNext->setEnabled(ilayer<layer.count()-1);
}


// swap current and previous layers so current layer tab moves down by one
bool CorteoParamWidget::moveLayerToPrevious() {
  int ilayer = tabWidgetLayers->currentIndex();
  if(ilayer<=0) return false;

  swapLayers(ilayer, ilayer-1);
  tabWidgetLayers->setCurrentIndex(ilayer-1);

  return true;
}


// swap current and next layers so current layer tab moves up by one
bool CorteoParamWidget::moveLayerToNext() {
  int ilayer = tabWidgetLayers->currentIndex();
  if(ilayer>=layer.count()-1) return false;

  swapLayers(ilayer, ilayer+1);
  tabWidgetLayers->setCurrentIndex(ilayer+1);

  return true;
}


void CorteoParamWidget::setlistFile() {
  listFile = QFileDialog::getSaveFileName(this,"Choose a file to store events in list mode.");
}


// set labels and enable widgets according to the selected technique
void CorteoParamWidget::update() {

  int tech = technique();

  // cone angle only for IBA
  labelConeAngle->setEnabled(tech>0);
  lineEditConeAngle->setEnabled(tech>0);

  // incident charge only for IBA
  lineEditCharge->setEnabled(tech != TECH_IMPLANT && tech != TECH_IMPREC);

  // energy discrimination enabled only for coincidence
  labelEnergyDiffDisc->setEnabled(tech == TECH_COINC);
  lineEditEnergyDiffDisc->setEnabled(tech == TECH_COINC);
  checkBoxLoadSeparateSpectra->setEnabled(tech > 0 || tech == TECH_IMPREC);

  // angle-to-detector-axis file only possible for IBA
  checkBoxGraphAngle->setEnabled(tech>0 || tech == TECH_FULLRBS || tech == TECH_FULLBACK );

  // detection toolbox tab
  toolBox->setItemEnabled(4,tech>0 || tech == TECH_FULLRBS || tech == TECH_FULLBACK );

  // disable all detector tabs then re-enable the useful ones
  for(int idet = 0; idet<detector.count(); idet++)
    tabWidgetDetectors->setTabEnabled(idet,false);

  switch(tech) {
  case TECH_IMPLANT :
  case TECH_IMPREC :
    // no detection
    break;
  case TECH_RBS_TOF :
    tabWidgetDetectors->setTabEnabled(2,true);
    // no break
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
    tabWidgetDetectors->setTabEnabled(0,true);
    tabWidgetDetectors->setCurrentIndex (0);
    break;
  case TECH_ERD_TOF :
    tabWidgetDetectors->setTabEnabled(3,true);
    // no break
  case TECH_ERD :
    tabWidgetDetectors->setTabEnabled(1,true);
    tabWidgetDetectors->setCurrentIndex (1);
    break;
  case TECH_COINC :
    tabWidgetDetectors->setTabEnabled(0,true);
    tabWidgetDetectors->setTabEnabled(1,true);
    tabWidgetDetectors->setCurrentIndex (0);
    break;
  default:
    QMessageBox::warning(this, "Corteo user interface", QString("Unknown technique code %1").arg(tech));
    break;
  }

  // cutoff angle possible only for RBS cross section
  labelCutoff->setEnabled(isBeamDetectionTechnique() && tech != TECH_FULLRBS);
  lineEditCutoff->setEnabled(isBeamDetectionTechnique() && tech != TECH_FULLRBS);

  // double-toggle checkboxes that enable other widgets
  checkBoxListMode->toggle();
  checkBoxListMode->toggle();

  // update layers according to technique
  for(int ilayer=0; ilayer<layer.count(); ilayer++) {
    bool b = layer[ilayer]->update(tech>0, tech==TECH_IMPREC);
    tabWidgetLayers->setTabIcon(ilayer, b?green:gray); // icon green or gray whether layer is used or not
    layer[ilayer]->labelNumber->setText(QString("Layer #%1").arg(ilayer));
  }

  // set charge label according to technique
  if(tech>0 || tech==TECH_FULLBACK || tech==TECH_FULLRBS)
      labelCharge->setText("Incident charge [C]:");
  else
      labelCharge->setText("Fluence [at/cm<sup>2</sup>]:");

  // calibration tabs

  // update used/unused layers in calibration forms compound correction tabs
  for(int i=0; i<calibration.count(); i++) {
    for(int ilayer=0; ilayer<layer.count(); ilayer++) {
      bool bl = layer[ilayer]->checkBoxUseLayer->isChecked();
      // icon green or gray whether layer is used or not
      calibration[i]->tabWidgetCompCorr->setTabIcon(ilayer, bl?green:gray);
      calibration[i]->setCompoundCorrEnabled(ilayer, bl);
    }
  }

  // remove current calibration tabs
  while(tabWidgetCalibrations->count())
    tabWidgetCalibrations->removeTab(0); // this doesn't destroy the forms themselves

  // beam calibration always there, at least the compound corrections...
  tabWidgetCalibrations->addTab(calibration[beamElement.Z], beamElement.symbol);
  // ... but enable its calibration entry fields only if we are doing spectrometry
  calibration[beamElement.Z]->setCalibrationFieldsEnabled(tech > 0 || tech == TECH_FULLRBS || tech == TECH_FULLBACK );

  // add tabs appropriate to technique considering currently used elements
  QList<int> alreadyThere;
  switch(tech) {
  case TECH_IMPLANT :
  case TECH_IMPREC :
    // no calibration
    break;
  case TECH_RBS_TOF :
  case TECH_ERD_TOF :
    tabWidgetCalibrations->addTab(calibration[0], "TOF");
    break;
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
  case TECH_COINC :
    tabWidgetCalibrations->addTab(calibration[beamElement.Z], beamElement.symbol);
    break;
  case TECH_ERD :
    // recoils
    for(int ilayer=0; ilayer<layer.count(); ilayer++) {
      if(layer[ilayer]->isUsed()) {
        for(int ielement=0; ielement<layer[ilayer]->element.count(); ielement++) {
          Atom at = layer[ilayer]->element[ielement]->element();
          // if and element is used and has to be simulated and it is not already in the calibration tabs, add it
          if(layer[ilayer]->element[ielement]->numberIsotopesToSimulate()>0 &&
             !alreadyThere.contains(at.Z)) {
            tabWidgetCalibrations->addTab(calibration[at.Z], at.symbol);
            alreadyThere.append(at.Z);
          }
        }
      }
    }
    break;
  default:
    QMessageBox::warning( this, "Corteo user interface", QString("Unknown technique code %1").arg(tech));
    break;
  }


  for(int i =0; i<detector.count(); i++)
    detector[i]->updateDetectorPoints();

}


// ask filename to load parameters from, and load them
void CorteoParamWidget::loadFile() {
  QString filename = QFileDialog::getOpenFileName(this,"Choose a Corteo User Interface parameter file",
    paramfile,"CorteoML file (*.cml)");

  if ( !filename.isEmpty() ) {
    paramfile = filename;
    loadParam( paramfile );
  }

}

// save parameters in current file unless it is not defined
void CorteoParamWidget::saveFile() {
  if(paramfile.isEmpty()) {
    saveFileAs();
    return;
  }
  saveParam( paramfile );
}

// ask filenam before saving parameters
void CorteoParamWidget::saveFileAs() {

  QString filename = QFileDialog::getSaveFileName(this,"Choose a filename to save under",
    paramfile,"Corteo user interface parameter file (*.cml)");

  if ( !filename.isEmpty() ) {
    paramfile = filename;
    saveParam( paramfile );
  }
}


// save parameters in default filename
bool CorteoParamWidget::saveParam() {

  return saveParam("CorteoUi.cml");

}


// lets follow the wave and use XML to save our parameters
// (it has the adventages of being structured, human editable and understandable,
//  and we can easily add or remove parameters)
bool CorteoParamWidget::saveParam(QString filename) {

  for(int i =0; i<detector.count(); i++)
    detector[i]->updateDetectorPoints();

  // create a CorteoML document from the parameters
  QDomDocument doc( "CorteoML" );
  doc.appendChild( XMLNode(doc) );

  // open file 'filename'
  QFile file( filename );
  if( !file.open( QIODevice::WriteOnly ) ) {
    QMessageBox::warning( 0, "Corteo user interface", "Can't open file " + filename );
    return false;
  }

  // wtite the CorteoML document to file
  QTextStream ts( &file );
  ts << doc.toString();

  file.close();
  return true;
}


// open CorteoML XML file to read parameters from
bool CorteoParamWidget::loadParam(QString filename) {

  QDomDocument doc( "CorteoML" );

  QFile file( filename );
  if( !file.open( QIODevice::ReadOnly ) )  {
    QMessageBox::warning( 0, "Corteo user interface", "Can't open parameter file " + filename );
    return false;
  }

  if( !doc.setContent( &file ) ) {
    file.close();
    QMessageBox::warning( 0, "Corteo user interface", "Can't find Corteo XML content in " + filename );
    return false;
  }

  QDomElement root = doc.documentElement();
  if( root.tagName() != "CorteoParam" )  {
    file.close();
    QMessageBox::warning( 0, "Corteo user interface", "Can't find <CorteoParam> tag in " + filename );
    return false;
  }

  setFromNode( root.firstChild() );

  file.close();
  update();
  return true;
}


QDomElement CorteoParamWidget::XMLNode(QDomDocument &doc) {

  QDomElement root = doc.createElement( "CorteoParam" );

  QDomElement deSimulation = doc.createElement( "simulation" );
  deSimulation.setAttribute( "technique",      comboBoxTechnique->currentText() );
  deSimulation.setAttribute( "number_proj",    lineEditNumberProjectiles->text() );
  deSimulation.setAttribute( "straggling",     stragglingType() );
  deSimulation.setAttribute( "mean_free_path", lineEditMeanFreePath->text() );
  deSimulation.setAttribute( "cone_angle",     lineEditConeAngle->text() );
  deSimulation.setAttribute( "minimum_energy", lineEditMinimumEnergy->text() );
  deSimulation.setAttribute( "cutoff",         lineEditCutoff->text() );
  deSimulation.setAttribute( "screening",      comboBoxScreening->currentIndex() );
  deSimulation.setAttribute( "nthread",        comboBoxNthread->currentIndex() );
  deSimulation.setAttribute( "random_seed",    lineEditRandomSeed->text() );
  root.appendChild( deSimulation );

  QDomElement deBeam = doc.createElement( "beam" );
  deBeam.setAttribute( "Z", beamElement.Z );
  deBeam.setAttribute( "selected_mass",   comboBoxBeamMass->currentIndex() );
  deBeam.setAttribute( "charge",          lineEditCharge->text() );
  deBeam.setAttribute( "energy",          lineEditEnergy->text() );
  deBeam.setAttribute( "angle_to_normal", lineEditAngleToNormal->text() );
  deBeam.setAttribute( "azimutal_angle",  lineEditAzimAngle->text() );
  deBeam.setAttribute( "diameter",        lineEditBeamDiameter->text() );
  root.appendChild( deBeam );

  QDomElement deLayers = doc.createElement( "layers" );
  deLayers.setAttribute( "current_layer", tabWidgetLayers->currentIndex() );
  for(int ilayer=0; ilayer<layer.count(); ilayer++) {
    deLayers.appendChild( layer[ilayer]->XMLNode(doc) );
  }
  root.appendChild( deLayers );

  QDomElement deDetectors = doc.createElement( "detectors" );
  deDetectors.setAttribute( "current_detector", tabWidgetDetectors->currentIndex() );
  deDetectors.setAttribute( "virtualWfactor",   lineEditWfactor->text()   );
  deDetectors.setAttribute( "virtualHfactor",   lineEditHfactor->text()   );
  for(int idet=0; idet<detector.count(); idet++)
    deDetectors.appendChild( detector[idet]->XMLNode(doc) );
  deDetectors.appendChild( foil->XMLNode(doc) );
  root.appendChild( deDetectors );

  QDomElement deSpectrum = doc.createElement( "spectrum" );
  deSpectrum.setAttribute( "number_channels", lineEditNumberChannels->text() );
  deSpectrum.setAttribute( "energy_disc",     lineEditEnergyDiffDisc->text() );
  deSpectrum.setAttribute( "list_mode",       checkBoxListMode->isChecked()?1:0 );
  deSpectrum.setAttribute( "list_file",       listFile );
  for(int i=0; i<calibration.count(); i++)
    deSpectrum.appendChild( calibration[i]->XMLNode(doc) );

  root.appendChild( deSpectrum );

  return root;

}


// set parameters from a series of XML nodes starting with n
void CorteoParamWidget::setFromNode( QDomNode n) {

  while( !n.isNull() ) {
    QDomElement e = n.toElement();
    if( !e.isNull() ) {
      if( e.tagName() == "simulation" ) {
        int index = comboBoxTechnique->findText(e.attribute("technique", techniqueName[0]));
        comboBoxTechnique->setCurrentIndex(index>=0?index:0);

        lineEditNumberProjectiles->setText( e.attribute("number_proj",     ""));
        setStragglingType(                  e.attribute("straggling",      "").toInt());
        lineEditMeanFreePath->setText(      e.attribute("mean_free_path",  ""));
        lineEditConeAngle->setText(         e.attribute("cone_angle",      ""));
        lineEditMinimumEnergy->setText(     e.attribute("minimum_energy",  ""));
        lineEditCutoff->setText(            e.attribute("cutoff",          ""));
        comboBoxScreening->setCurrentIndex( e.attribute("screening",       "").toInt() );
        comboBoxNthread->setCurrentIndex(   e.attribute("nthread",         "").toInt() );
        lineEditRandomSeed->setText(        e.attribute("random_seed",  "8550983"));
        tabWidgetDetectors->setCurrentIndex(e.attribute("current_detector","").toInt() );

      } else if( e.tagName() == "beam" ) {

        beamElement = ElementForm::setElement(e.attribute("Z","").toInt(), pushButtonBeam, comboBoxBeamMass);
        comboBoxBeamMass->setCurrentIndex( e.attribute("selected_mass",  "").toInt());
        lineEditCharge->setText(           e.attribute("charge",         ""));
        lineEditEnergy->setText(           e.attribute("energy",         ""));
        lineEditAngleToNormal->setText(    e.attribute("angle_to_normal",""));
        lineEditAzimAngle->setText(        e.attribute("azimutal_angle", ""));
        lineEditBeamDiameter->setText(     e.attribute("diameter",       ""));

      } else if( e.tagName() == "layers" ) {

        QDomNode nn = n.firstChild();
        int ilayer = 0;
        while( !nn.isNull() ) {
          if(ilayer>=layer.count()) {
            QMessageBox::warning( 0, "Corteo user interface",
              "Too many layers in parameter file, only loading first "+QString::number(layer.count()) );
            break;
          } else {
            layer[ilayer]->setFromNode(nn);
          }
          nn = nn.nextSibling();
          ilayer++;
        }
        tabWidgetLayers->setCurrentIndex(e.attribute("current_layer", "").toInt());

      } else if( e.tagName() == "detectors" ) {
        tabWidgetDetectors->setCurrentIndex(e.attribute("current_detector","").toInt() );
        lineEditWfactor->setText(           e.attribute("virtualWfactor", ""));
        lineEditHfactor->setText(           e.attribute("virtualHfactor", ""));

        QDomNode nn = n.firstChild();
        int idet = 0;
        while( !nn.isNull() ) {
          if(nn.toElement().tagName()=="layer") {
            foil->setFromNode(nn);
          } else if(idet<detector.count()) {
            detector[idet]->setFromNode(nn);
            idet++;
          } else {
            QMessageBox::warning( 0, "Corteo user interface", "Too many detectors in parameter file, only loading first "+QString::number(detector.count()) );
            break;
          }
          nn = nn.nextSibling();
        }
      } else if( e.tagName() == "spectrum" ) {
        lineEditNumberChannels->setText( e.attribute("number_channels", "1000"));
        lineEditEnergyDiffDisc->setText( e.attribute("energy_disc", "1000"));
        checkBoxListMode->setChecked(   (e.attribute("list_mode",  "").toInt()==1)?true:false );
        listFile =                               e.attribute("list_file",  "");

        // read calibration for each element we have at this point in our 'projectiles' list
        QDomNode nn = n.firstChild();
        while( !nn.isNull() ) {
          QDomElement ee = nn.toElement();
          if( !ee.isNull() && ee.tagName() == "calibration" ) {

            // check if ee corresponds to an element in our list
            int Z = ee.attribute("Z", "").toInt();
            if(Z>=0 && Z < calibration.count()) {
              calibration[Z]->setFromNode(nn);
            }
          }
          nn = nn.nextSibling();
        }
      }
    }
    n = n.nextSibling();
  }
}


// Generate the content of a corteo input file as a string
// On error, sets 'error' true and returns an error message describing the problem
QString CorteoParamWidget::corteo(bool * error, int tech) {

  *error = false;

  QString scorteo;
  QTextStream ts( &scorteo );

  QString serror = "Can't run a Corteo simulation because the following parameters are not properly set:\n";

  // general parameters: technique, mean free path, cone angle, minimum energy, energy stragling, writeDetails, screening
  if(lineEditMeanFreePath->text().isEmpty() ) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="mean free path";
  }
  if(lineEditConeAngle->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="cone angle";
  }
  if(lineEditMinimumEnergy->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="minimum energy";
  }
  if(lineEditRandomSeed->text().isEmpty() || lineEditRandomSeed->text().toInt()<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="random seed";
  }

  ts << "% technique, mean free path, cone angle [deg], minimum energy [eV], energy stragling, writeDetails, screening, cutoff angle, random seed, nthreads" << endl;
  ts << tech                                   << " "
     << lineEditMeanFreePath->text()           << " "
     << lineEditConeAngle->text()              << " "
     << lineEditMinimumEnergy->text()          << " "
     << stragglingType()                       << " ";

  // details
  int writeDetails = 0;
  // no longer used in corteo, writeDetails != 0 now just produces (a lot) more output, for specialists!
  // if(tech>0 && checkBoxGraphAngle->isChecked()) writeDetails += 2; // for IBA, write angle-to-detector-axis data
  // if(tech>0 && checkBoxListMode->isChecked()) writeDetails += 4; // for IBA, write detected atoms detection details
  // if(tech<0 && checkBoxListMode->isChecked()) writeDetails += 1; // for implantation, write where each ion stopped
  ts << writeDetails;

  // screening
  ts << " " << screening();

  // collision cutoff angle (only for RBS)
  if( isBeamDetectionTechnique() && !lineEditCutoff->text().isEmpty() )
    ts << " " << lineEditCutoff->text();
  else
    ts << " 0";

  //number of parallel threads
  ts << " " <<  lineEditRandomSeed->text().toInt();

  //number of parallel threads
  ts << " " << nthread()  << endl;

  // beam parameters: Z, atomic mass, number of ions, beam energy, beam radius, directional cosines
  if(lineEditNumberProjectiles->text().isEmpty() ||lineEditNumberProjectiles->text().toInt()<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="number of projectiles";
  }

  if(lineEditEnergy->text().isEmpty() || lineEditEnergy->text().toDouble()<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam energy";
  }
  if(lineEditEnergy->text().toDouble()<=lineEditMinimumEnergy->text().toDouble()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam energy smaller or equal to minimum energy";
  }
  if(lineEditBeamDiameter->text().isEmpty() || lineEditBeamDiameter->text().toDouble()<0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam diameter";
  }
  if(lineEditAngleToNormal->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam angle to normal";
  }
  if(lineEditAzimAngle->text().isEmpty()) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="beam azimutal angle";
  }

  int numproj = lineEditNumberProjectiles->text().toInt();

  ts << "% beam Z, atomic mass, number of ions, beam energy [eV], beam radius [A]" << endl
     << beamElement.Z                      << " "
     << comboBoxBeamMass->currentText()    << " ";
  if(tech>0) ts << 0                       << " "; // to avoid confusion, put 0 projectiles here, and detail them for each layer
  else  ts << numproj                      << " "; // for implantation and "Full RBS", set it here
  ts << lineEditEnergy->text().toDouble()  << " "
     << lineEditBeamDiameter->text().toDouble()*0.5e7 << endl;  // divide by to 2 to get radius and mult. by 1e7 to get A

  // directionnal cosines
  double theta = lineEditAngleToNormal->text().toDouble()*PI/180;
  double phy   = lineEditAzimAngle->text().toDouble()*PI/180;
  ts << "% incident beam directional cosines"  << endl
     << cos(theta) << " " << sin(theta)*cos(phy) << " " << sin(theta)*sin(phy) << endl;


  double targetThickness = 0.;
  for(int i=0; i<layer.count(); i++)
    if(layer[i]->isUsed() )
      targetThickness += layer[i]->lineEditThickness->text().toDouble();
  if(targetThickness<=0) {
    if(*error) serror+= ", ";
    else *error = true;
    serror+="target has no thickness";
  }

  bool error2;
  QString ss;

  // count number of used layers and print it
  int numberLayers=0;
  for(int i=0; i<layer.count(); i++)
    if(layer[i]->isUsed())
      numberLayers++;

  if(numberLayers==0) {
    // no used layer
    if(*error) serror+= ", ";
    else *error = true;
    serror+="target has no layers";
  }

  ts << "% number of layers, number of traveling atoms" << endl;
  ts << numberLayers << " " << tabWidgetCalibrations->count()  << endl;

  // layers: thickness, density, numElements
  for(int i=0; i<layer.count(); i++) {
    if(layer[i]->isUsed()) {
      ts << "% layer thickness, roughness, atomic density, number of elements, number of projectiles" << endl;
      ss = layer[i]->corteo(QString("layer %1").arg(i), numproj, &error2 );
      if(error2) {
        serror+=ss;
        *error = true;
      } else {
        ts << ss << endl;
      }
    }
  }

  // foil/absorber layer
  if(foil->isUsed()) {
    ss = foil->corteo(QString("foil"), 0,  &error2);
    if(error2) {
      serror+=ss;
      *error = true;
    } else {
      ts << ss << endl;
    }
  } else {
    ts << "0 0 0 1 1 1\n 1 1 1 1 1" << endl; // no absorber, we just need a dummy layer of thickness 0
  }

  // detectors
  for(int i =0; i<detector.count(); i++)
    detector[i]->updateDetectorPoints();

  // find which detector we are using
  int firstDetector = -1, secondDetector = -1;
  switch(tech) {
  case TECH_IMPLANT:
  case TECH_IMPREC :
    break;
  case TECH_FULLRBS :
  case TECH_RBS :
    firstDetector = 0;
    break;
  case TECH_RBS_TOF :
    firstDetector = 0;
    secondDetector = 2;
    break;
  case TECH_ERD :
    firstDetector = 1;
    break;
  case TECH_ERD_TOF :
    firstDetector = 1;
    secondDetector = 3;
    break;
  case TECH_COINC :
    firstDetector = 0;
    secondDetector = 1;
    break;
  default:
    QMessageBox::warning( 0, "Corteo user interface", QString("Technique code %1 unknown.").arg(tech));
    break;
  }

  // write detector information or, if detector not used, dummy information so corteo is happy
  ts << "% detectors: isEllipse p1x p1y p1z p2x p2y p2z p3x p3y p3z r_annulus" << endl;
  QString dummyDetector = "0  -1 0 0  -1 1 0  -1 1 -1  0";
  if(firstDetector == -1) {
    ts << dummyDetector << endl;
  } else {
    ss = detector[firstDetector]->corteo(&error2);
    if(error2) {
      serror+=ss;
      *error = true;
    } else {
      ts << ss << endl;
    }
  }

  if(secondDetector == -1) {
    ts << dummyDetector << endl;
  } else {
    ss = detector[secondDetector]->corteo(&error2);
    if(error2) {
      serror+=ss;
      *error = true;
    } else {
      ts << ss << endl;
    }
  }

  ts << "% virtual detector enlargement factors: width, height" << endl;
  QString Ws = lineEditWfactor->text().isEmpty()?QString("1"):lineEditWfactor->text(); // default is 1
  QString Hs = lineEditHfactor->text().isEmpty()?QString("1"):lineEditHfactor->text(); // default is 1
  if(Ws.toDouble()<1.0 || Hs.toDouble()<1.0) {
    if(*error) serror += ", ";
    else *error = true;
    serror += "virtual factors can't be smaller than 1";
  }
  ts << Ws << " " << Hs << endl;

  // compound correction of each detected/traveling ion in each layer
  // (first data is Z, then compound corr. in each used layer)
  ts << "% ion Z, and its compound correction in each layer" << endl;
  for(int ielem=0; ielem<tabWidgetCalibrations->count(); ielem++) {
      CalibrationForm *calib = (CalibrationForm *) tabWidgetCalibrations->widget(ielem);
      if(ielem>0) ts << " ";
      ts << calib->getZ();
      for(int i=0; i<layer.count(); i++)
          if(layer[i]->isUsed())
              ts << " " << calib->compoundCorr(i);
  }

  if(*error) return serror;
  return scorteo;

}


// fire up a foil edit dialog
void CorteoParamWidget::editFoil() {
  foilDialog->exec();
}

// return true if layer is used (and within allowed number of layers)
bool CorteoParamWidget::isUsedLayer(int ilayer) {
  if(ilayer<0 || ilayer>=layer.count()) return false;
  return layer[ilayer]->isUsed();
}

// return true if element and layer are used (and within allowed numbers)
bool CorteoParamWidget::isUsedElement(int ilayer, int ielement) {
  if(!isUsedLayer(ilayer)) return false;
  if(ielement<0 || ielement>=layer[ilayer]->element.count()) return false;
  return layer[ilayer]->element[ielement]->isUsed();
}

// return true if isUsedElement is true and element is selected for simulation
bool CorteoParamWidget::isSimulatedElement(int ilayer, int ielement) {
  if(isUsedElement(ilayer, ielement) && layer[ilayer]->element[ielement]->checkBoxSimulate->isChecked()) return true;
  return false;
}


// return the number of isotopes to simulate for an element in a layer
// return 0 if isSimulatedElement returns false
// return 1 if checkBoxIsotopes is not checked
// otherwise return the number of isotopes with abundance > 1%
int  CorteoParamWidget::numberIsotopesToSimulate(int ilayer, int ielement) {
  if(!isSimulatedElement(ilayer, ielement)) return 0;
  // the previous condition also checked the validity of ilayer, ielement arguments through a call to isUsedElement
  // so we can safely access this element form without further checking
  return layer[ilayer]->element[ielement]->numberIsotopesToSimulate();

}


// return atom structure corresponfing to element ielement in layer ilayer
// (return null atom if ielement or ilayer are out of bounds)
Atom CorteoParamWidget::element(int ilayer, int ielement) {
  if(ilayer<0   || ilayer>=layer.count()) return Atom();
  if(ielement<0 || ielement>=layer[ilayer]->element.count()) return Atom();
  return layer[ilayer]->element[ielement]->element();
}

// fraction of ielement in layer
double CorteoParamWidget::fraction(int ilayer, int ielement) {
  if(ilayer<0   || ilayer>=layer.count()) return 0;
  if(ielement<0 || ielement>=layer[ilayer]->element.count()) return 0;

  double sum = 0;
  for(int i=0; i<layer[ilayer]->element.count(); i++)
    sum += layer[ilayer]->element[i]->concentration();

  if(sum>0) return layer[ilayer]->element[ielement]->concentration()/sum;
  return 0;
}

// number of element in layer ilayer
int CorteoParamWidget::numberElement(int ilayer) {
  if(ilayer<0 || ilayer>=layer.count()) return -1;
  return layer[ilayer]->element.count();
}

// number of layers
int CorteoParamWidget::numberLayer() {
  return layer.count();
}

// currently selected technique
int CorteoParamWidget::technique() {
  return techniqueCode[comboBoxTechnique->currentIndex()];
}

// currently selected screening
int CorteoParamWidget::screening() {
  return comboBoxScreening->currentIndex();
}

// currently selected screening
int CorteoParamWidget::nthread() {
  return comboBoxNthread->currentText().toInt();
}


// is the detected particle corresponding to a simulated element is actually the beam?
bool CorteoParamWidget::isBeamDetectionTechnique () {
  switch(technique()) {
  case TECH_IMPLANT:
  case TECH_IMPREC :
  case TECH_COINC :
  case TECH_ERD :
  case TECH_ERD_TOF :
    return false;
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
  case TECH_RBS_TOF :
    return true;
  default:
    QMessageBox::warning( 0, "Corteo user interface", QString("Technique code %1 unknown.").arg(technique()));
    break;
  }
  return false;
}

// is the detected particle corresponding to a simulated element is actually the beam?
bool CorteoParamWidget::isTOF () {
  switch(technique()) {
  case TECH_IMPLANT:
  case TECH_IMPREC :
  case TECH_COINC :
  case TECH_ERD :
  case TECH_RBS :
  case TECH_FULLRBS :
  case TECH_FULLBACK :
    return false;
  case TECH_ERD_TOF :
  case TECH_RBS_TOF :
    return true;
  default:
    QMessageBox::warning( 0, "Corteo user interface", QString("Technique code %1 unknown.").arg(technique()));
    break;
  }
  return false;
}


void CorteoParamWidget::showCorteoLog() {

  QFile file("CorteoUi.log");
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text)){
    QMessageBox::warning( 0, "Corteo user interface", "Can't read 'CorteuUi.log'.");
    return;
  }

  QTextStream in(&file);
  QDialog * showLogDialog = new QDialog(this);
  showLogDialog->resize(600,500);

  // create a QTextEdit that contains the content of CorteuUi.log
  // ans whos parent is showLogDialog
  QTextEdit * te = new QTextEdit();

  QHBoxLayout *layout = new QHBoxLayout(showLogDialog);
  layout->addWidget(te);

  te->setLineWrapMode(QTextEdit::NoWrap);
  te->insertPlainText(in.readAll());
  showLogDialog->show();
  te->ensureCursorVisible();  // to make the QTextEdit scroll to the end

  showLogDialog->exec();
  delete showLogDialog;
}

int CorteoParamWidget::numberSimul() {
    int numberSim = 0;
    for(int ilayer=0; ilayer<numberLayer(); ilayer++)
        for(int ielement=0; ielement<numberElement(ilayer); ielement++)
            if(isSimulatedElement(ilayer, ielement))
                numberSim+=numberIsotopesToSimulate(ilayer, ielement);

    return numberSim;

}

void CorteoParamWidget::aboutCorteoUi() {
  QSplashScreen *splash = new QSplashScreen(QPixmap(":/images/splash.png") ) ;
  splash->setWindowFlags(Qt::Tool) ;
  splash->setWindowTitle("About Corteo user interface") ;
  splash->show() ;
}

