/* This file is part of the Allegria project
   Copyright (C) 2004 Francois Schiettekatte <francois.schiettekatte@umontreal.ca>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifndef FSGRAPH_H
#define FSGRAPH_H

#include <QPainter>
#include <QWidget>
#include <QApplication>
#include <QMouseEvent>

#include <math.h>

#include "FSAxis.h"
#include "FSDataSet.h"
#include "FSMarker.h"

class FSGraph : public QWidget
{
    Q_OBJECT
public:

    FSAxis* x;
    FSAxis* y;

    bool                readOnlyGraph; // used to disable many commands
	bool				allowSpots;    // allow user to define spots using Shift + left click

    QFont               legendFont;
    QPen                legendPen;

    FSDataSet *         selectedDataSet;
    FSDataSet *         firstDataSet;  // pointer to the first dataset (pointer to the next is found in firstDataSet and so on)

    FSDataSet *         spots;  // list of spots marked by the user (Shift + left click)

private:
    QString             graphName;
  	QString				confFile;    // file where to store graph parameters

	// after a call to drawIt()
    int xLeg;           // x coordinate of the widest idem of the legend box
    int yLeg;           // y coordinate of the highest item in legend box
    int legendWidth;    // width of the widest item (excuding margin of lenght 'fs/2')
    int legendHeight;   // height of all items (excuding margin of lenght 'fs/2')
    int legendSpacing;  // spacing between legend items (can be smaller than fs/2 if we need to skweeze the legend)

	bool				hasChanged;	// should we repaint or just re-display the graph
	QPixmap				pm;			// pixmap to which the graph is drawn for re-display
	QSize				previousSize; // to check if window size has changed

    bool                _isZoomed;
    bool                _isZooming;
    QRect               _zoomRect;

    int                 ndatasets;

    char                legendPos;        // 0: none, 1: inside left, 2:inside right 3:outside right
    double              fsize;         // size factor

    //QRect               _selectRect;

    FSMarker *          firstMarker;   // marked area on which operations occure
    int                  nmarkers;
    bool                _isMarking;
    QRect               _markRect;

// class constructor
public:
    FSGraph(QWidget * parent=0);
    FSGraph(QString conffile, QWidget * parent=0);
   ~FSGraph();

public slots:
    void clearMarkers( void );
    void loadFile();
    void saveSelectedDataSet();
    void unZoom( void );

signals:
	void spotCreated();

private slots:
    int  deleteSelectedDataSet( void );
    void copySelectedDataSet( void );
    void clipboardSelectedDataSet( void );
    void clipboardGraph(void);
    void sortSelectedDataSet( void );
    void swapSelectedDataSet( void );

    // dialogs
    void editLegend( void );
    void editSelectedDataSet( void );
    void editYAxis( void );
    void editXAxis( void );
    void toggleYAxis( void );
    void toggleXAxis( void );
    void slotOp( void );
    void fit( void );

// Methods
public:
    void loadFile( bool clearAllDatasets );
    void clearDataSets();
    void appendDataSet(FSDataSet* ddataset, bool differentColor = false);
    void appendMarker(double x1, double y1, double x2, double y2);
    void updateGraph();
    void zoom(double xmin, double xmax, double ymin, double ymax);

protected:
    void drawIt( QPainter * );
    void paintEvent( QPaintEvent * );
    void mousePressEvent( QMouseEvent * e );
    void mouseMoveEvent( QMouseEvent * e );
    void mouseReleaseEvent( QMouseEvent * e );
    void keyPressEvent( QKeyEvent * e );
    void keyReleaseEvent( QKeyEvent * e );

    bool legendSize(QPainter *dc, QRect r);
    void operation(int op=0);

private:
	void init();
    void majorant();
    void drawMarkers(QPainter * dc , QRect r);
    int  isInsideMarker(bool *insideMark);

// properties
public:
    FSDataSet * yProjectionOfSelectedDataset(double binSize=1);
    QString getGraphName( void ) { return graphName; }
    void   setGraphName( QString s ) { graphName = s; setWindowTitle(graphName); }

    int    getndatasets( void ) { return ndatasets; }
    char   getlegendPos( void ) { return legendPos; }
    bool   setlegendPos( char n ) { if(n>=0&&n<=3) legendPos=n; return (n>=0&&n<=3); }
    double getfsize( void ) { return fsize; }
    double setfsize( double n ) { if(n>0.) fsize=n; return n>0.; }

};

#endif // FSGRAPH_H
